#include <stdio.h>
#include <pthread.h>
#include "prover.h"
#include "uthash.h"
#include "tree.h"
#include "symbol_table.h"
#include "clauses.h"

extern int improved_snf_plus;

extern int ml_prover;
extern int modal_negative;
extern int strong_modal_negative;
extern int numclause;
extern int newtemp;
extern int numdeleted;

extern hml_mclauses *m_usable;
extern hml_mclauses *m_sos;
extern hml_pclauses *l_sos;

extern int numprops;

extern void print_sos (void);
extern void delete_clause_all (int number);
extern tclause *get_clause (int clause);
extern prop_node *find_prop(int p);
extern int rename_formula(tnode *formula, int type, int set, int subtype, int modal_level);
extern void add_definition(int id, tnode *formula, int set, int subtype, int modal_level);
extern tnode *create_tnode(int type, int id, int mdepth, tnode *left, tnode *right, formulalist *list);
extern void insert_clause_deleted (tclause *clause);
extern void insert_clause_prop (tclause **clause, hml_pclauses **set);
extern void insert_clause_modal (tclause **clause, hml_mclauses **set);
extern clauseslist *delete_clause_list(tclause *c, clauseslist *l);
extern int classify_clause (literalslist *l);
extern prop_node *insert_pnew_node (int id, int type);

extern int check_minimal(tclause *clause, int modal_level);

typedef struct clauseslistsnfminus {
  tclause *clause;
  struct clauseslistsnfminus *next;
} clauseslistsnfminus;


/* Scans the modal formulae doing the rewriting and renaming when needed */


void free_clauses_set_m_snf_minus (hml_mclauses **set) {
  hml_mclauses *aux, *auxtmp;
  hagentclauses *hag, *hagtmp;
  htypeclauses *htc, *htctmp;
  hmaxlitclauses *hmaxlit, *hmaxlittmp;

  aux = *set;

  HASH_ITER(hml,*set,aux,auxtmp) {
    HASH_ITER(hid,aux->id,hag,hagtmp) {
      HASH_ITER(htype,hag->type,htc,htctmp) {
	HASH_ITER(hmax,htc->lit,hmaxlit,hmaxlittmp) {
	  if ((hmaxlit->max != CTRUE && hmaxlit->max != CFALSE) && (strong_modal_negative || modal_negative)) {
	    clauseslist *list = hmaxlit->list;
	    while (list != NULL) {
	      tclause *clause = get_clause(list->clause_number);
	      list = list->next;
	      if (clause->deleted != NULL) {
		hmaxlit->list = delete_clause_list(clause,hmaxlit->list);
		delete_clause_all(clause->number);
		numdeleted++;
		insert_clause_deleted(clause);
	      }
	    }
	  }
	  if (hmaxlit->list == NULL) {
	    HASH_DELETE(hmax,htc->lit,hmaxlit);
	    free(hmaxlit);
	    hmaxlit = NULL;
	  }
	}
	if (htc->lit == NULL) {
	  HASH_DELETE(htype,hag->type,htc);
	  free(htc);
	  htc = NULL;
	}
      }
      if (hag->type == NULL) {
	HASH_DELETE(hid,aux->id,hag);
	free(hag);
	hag = NULL;
      }
    }
    if (aux == NULL) {
      HASH_DELETE(hml,*set,aux);
      free(aux);
      aux = NULL;
    }
  }
}


void snf_minus_by_set (hml_mclauses **modal_set) {

  hml_mclauses *aux;
  hagentclauses *hag = NULL;
  htypeclauses *htc = NULL;
  hmaxlitclauses *hmaxlit = NULL;

  clauseslistsnfminus *newclauses = NULL;
  
  for (aux=*modal_set; aux != NULL; aux=aux->hml.next)
      for (hag=aux->id; hag != NULL; hag=hag->hid.next)
	for (htc=hag->type; htc != NULL; htc=htc->htype.next)
	  for (hmaxlit = htc->lit; hmaxlit != NULL; hmaxlit=hmaxlit->hmax.next) {
	    if ((hmaxlit->max != CTRUE && hmaxlit->max != CFALSE) && (strong_modal_negative || (modal_negative && hmaxlit->max > 0))) {
	      clauseslist *l = hmaxlit->list;
	      while (l != NULL) {
		tclause *clause = get_clause(l->clause_number);

		int minimal = 0;
		if (hmaxlit->max < 0 && improved_snf_plus) {
		  minimal = check_minimal(clause,clause->modal_level);
		}

		if (!minimal) {
		  int ml;
		  if (ml_prover) {
		    ml = clause->modal_level + 1;
		  }
		  else ml = clause->modal_level;
  
		  /* Renaming the right-hand side */

		  prop_node *p = NULL;
		  int id_minus_right;
		  if (clause->max_literal > 0) {
		    p = find_prop(clause->max_literal);
		    if (p->rennegby == 0) {
		      prop_node *q = insert_pnew_node(newtemp++,HIGH);
		      p->rennegby = q->id;
		    }
		    id_minus_right = -p->rennegby;
		  } else id_minus_right = clause->max_literal;

		  /* Renaming the left-hand side. This needs to be improved. */

		  int id_minus_left;
		  if (clause->left != CTRUE && clause->left > 0) {
		    p = find_prop(clause->left);
		    if (p->renposby == 0) {
		      prop_node *q = insert_pnew_node(newtemp++,HIGH);
		      p->renposby = q->id;
		    }
		    id_minus_left = -p->renposby;
		  } else id_minus_left = clause->left;

		  /* Adding the new modal clause */

		  if ((clause->left != CTRUE && clause->left > 0) || clause->max_literal > 0) {
		    tclause *newclause = malloc(sizeof(tclause));
		    newclause->number = numclause++;
		    newclause->type = clause->type;
		    newclause->subtype = clause->subtype;
		    newclause->numdiamond = clause->numdiamond;
		    newclause->size = clause->size;
		    newclause->modal_level = clause->modal_level;
		    newclause->max_literal = id_minus_right;
		    newclause->min_literal = id_minus_right;
		    newclause->agent = clause->agent;
		    newclause->class = clause->class;
		    newclause->propagated = 0;
		    newclause->left = id_minus_left;
		    newclause->right = malloc(sizeof(literalslist));
		    newclause->right->literal = id_minus_right;
		    newclause->right->next = NULL;
		    newclause->just = malloc(sizeof(justification));
		    newclause->just->rule = SNF_MINUS;
		    clauseslist *parent = malloc(sizeof(clauseslist));
		    parent->clause_number = clause->number;
		    parent->next = NULL;	  
		    newclause->just->parents = parent;
		    newclause->just->literals = NULL;
		    newclause->deleted = NULL;

		    clauseslist *parent_deleted = malloc(sizeof(clauseslist));
		    parent_deleted->clause_number = newclause->number;
		    parent_deleted->next = NULL;
      
		    literalslist *literals = malloc(sizeof(literalslist));
		    literals->literal = clause->right->literal;
		    literals->next = NULL;

		    justification *deleted = malloc(sizeof(justification));
		    deleted->rule = SNF_MINUS;
		    deleted->parents = parent_deleted;
		    deleted->literals = literals;
		  
		    clause->deleted = deleted;

		    clauseslistsnfminus *auxclauses = malloc(sizeof(clauseslistsnfminus));
		    auxclauses->clause = newclause;
		    auxclauses->next = newclauses;
		    newclauses = auxclauses;
		  }

		  /* Adding the new literal clause, if the literal on the rhs is renamed. */

		  if (clause->max_literal > 0) {
		    tclause *literal = malloc(sizeof(tclause));
		    literal->number = numclause++;
		    literal->type = LITERAL;
		    literal->subtype = clause->subtype;
		    literal->numdiamond = clause->numdiamond;
		    literal->size = 2;
		    literal->modal_level = ml;
		    literal->max_literal = clause->max_literal;
		    literal->min_literal = -id_minus_right;
		    literal->agent = 0;
		    literal->left = CTRUE;
		    literal->propagated = 0;
		    literal->right = malloc(sizeof(literalslist));
		    literal->right->literal = -id_minus_right;
		    literal->right->next = malloc(sizeof(literalslist));
		    literal->right->next->literal = clause->max_literal;
		    literal->right->next->next = NULL;
		    literal->class = classify_clause(literal->right);
		    literal->just = malloc(sizeof(justification));
		    literal->just->rule = SNF_MINUS;
		    clauseslist *parentl = malloc(sizeof(clauseslist));
		    parentl->clause_number = clause->number;
		    parentl->next = NULL;	  
		    literal->just->parents = parentl;
		    literal->just->literals = NULL;
		    literal->deleted = NULL;
		    insert_clause_prop(&literal,&l_sos);
		  }


		  /* Adding the new literal clause, if the literal on the lhs is renamed. */

		  if (clause->left > CTRUE) {
		    tclause *literal = malloc(sizeof(tclause));
		    literal->number = numclause++;
		    literal->type = LITERAL;
		    literal->subtype = clause->subtype;
		    literal->numdiamond = clause->numdiamond;
		    literal->size = 2;
		    literal->modal_level = clause->modal_level;
		    literal->max_literal = clause->left;
		    literal->min_literal = id_minus_left;
		    literal->agent = 0;
		    literal->left = CTRUE;
		    literal->propagated = 0;
		    literal->right = malloc(sizeof(literalslist));
		    literal->right->literal = id_minus_left;
		    literal->right->next = malloc(sizeof(literalslist));
		    literal->right->next->literal = -clause->left;
		    literal->right->next->next = NULL;
		    literal->class = classify_clause(literal->right);
		    literal->just = malloc(sizeof(justification));
		    literal->just->rule = SNF_MINUS;
		    clauseslist *parentl = malloc(sizeof(clauseslist));
		    parentl->clause_number = clause->number;
		    parentl->next = NULL;	  
		    literal->just->parents = parentl;
		    literal->just->literals = NULL;
		    literal->deleted = NULL;
		    insert_clause_prop(&literal,&l_sos);
		  }
		  
		}
		l = l->next;	
	      }
	    }
	  }
  
  free_clauses_set_m_snf_minus(&(*modal_set));

  while (newclauses != NULL) {
    insert_clause_modal(&(newclauses->clause),&(*modal_set));
    clauseslistsnfminus *auxclauses = newclauses->next;
    free(newclauses);
    newclauses = auxclauses;
  }

}

void snf_minus(void) {
  snf_minus_by_set(&m_usable);
  snf_minus_by_set(&m_sos);
}
